#include "packages.h"

#include "runtime/threading.h"
#include "runtime/assets.h"

#include <limits.h>
#include <stdlib.h>
#include <string.h>

typedef struct {
    vy_uid uid;
    void *buffer;
    size_t buffer_size;
} vy_package_entry;

typedef struct {
    char *name;
    unsigned int num_entries;
    unsigned int entry_capacity;
    vy_package_entry *entries;
} vy_package;

#define MAX_PACKAGES 1024

vy_package _packages[MAX_PACKAGES];
unsigned int _package_count = 0;

vy_mutex *_guard;

unsigned int vyAddPackageFile(vy_text_span name) {
    vyLockMutex(_guard);

    for (unsigned int i = 0; i < _package_count; ++i) {
        if (vyCompareSpanToString(name, _packages[i].name) == 0) {
            vyUnlockMutex(_guard);
            return i;
        }
    }

    /* Create a new package */
    _packages[_package_count].name = malloc(name.length + 1);
    if (!_packages[_package_count].name) {
        vyReportError("ASSETC",
                      "Failed to allocate storage for new package %*.s",
                      name.length,
                      name.start);
        vyUnlockMutex(_guard);
        return UINT_MAX;
    }
    memcpy(_packages[_package_count].name, name.start, name.length);
    _packages[_package_count].name[name.length] = '\0';

    unsigned int index = _package_count++;

    vyUnlockMutex(_guard);
    return index;
}

void vyInitPackages(void) {
    _guard = vyCreateMutex();
    /* Create the default package (0) */
    vyAddPackageFile((vy_text_span){.start = "default.pkg", .length = 12});
}

void vyAddAssetToPackage(unsigned int package, vy_uid uid, void *buffer, size_t size) {
    vyLockMutex(_guard);
    if (package >= _package_count) {
        vyReportError("ASSETC", "Trying to add an asset to a non-existing package.");
        vyUnlockMutex(_guard);
        return;
    }

    vy_package *pkg = &_packages[package];
    if (pkg->num_entries == pkg->entry_capacity) {
        unsigned int new_cap = (pkg->entry_capacity > 0) ? 2 * pkg->entry_capacity : 256;
        vy_package_entry *n  = realloc(pkg->entries, new_cap * sizeof(vy_package_entry));
        if (!n) {
            vyReportError("ASSETC", "Failed to grow storage for package %u.", package);
            return;
        }
        pkg->entry_capacity = new_cap;
        pkg->entries        = n;
    }
    pkg->entries[pkg->num_entries].buffer = buffer;
    pkg->entries[pkg->num_entries].buffer_size = size;
    pkg->entries[pkg->num_entries].uid = uid;
    ++pkg->num_entries;
    
    vyUnlockMutex(_guard);
}